from data_gateway_sdk.protocols.smpm.smpm_dl_device_energy_8b_get_data import SmpmDlDeviceEnergy8BGetDataData, SmpmDlDeviceEnergy8bGetDataDataRequestMonth, \
    SmpmDlDeviceEnergy8bGetDataDataRequestId
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_dl_device_energy_8b_get_data() -> None:
    case_serialized = bytes.fromhex("8001c5930c780300")
    assert SmpmDlDeviceEnergy8BGetDataData(year=2032, month=SmpmDlDeviceEnergy8bGetDataDataRequestMonth.JAN, day=15, request_data_pack_ids=(SmpmDlDeviceEnergy8bGetDataDataRequestId.DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_2, SmpmDlDeviceEnergy8bGetDataDataRequestId.NETWORK_PARAMS_PHASE1)) == SmpmDlDeviceEnergy8BGetDataData.parse(BufRef(case_serialized))  # noqa: E501
    case_serialized = bytes.fromhex("800130900c780300")  # overflow
    assert SmpmDlDeviceEnergy8BGetDataData(year=2000, month=SmpmDlDeviceEnergy8bGetDataDataRequestMonth.DEC, day=0, request_data_pack_ids=(SmpmDlDeviceEnergy8bGetDataDataRequestId.DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_2, SmpmDlDeviceEnergy8bGetDataDataRequestId.NETWORK_PARAMS_PHASE1)) == SmpmDlDeviceEnergy8BGetDataData.parse(BufRef(case_serialized))  # noqa: E501
    case_serialized = bytes.fromhex("80f933900c780300")  # overflow
    assert SmpmDlDeviceEnergy8BGetDataData(year=2127, month=SmpmDlDeviceEnergy8bGetDataDataRequestMonth.DEC, day=0, request_data_pack_ids=(SmpmDlDeviceEnergy8bGetDataDataRequestId.DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_2, SmpmDlDeviceEnergy8bGetDataDataRequestId.NETWORK_PARAMS_PHASE1)) == SmpmDlDeviceEnergy8BGetDataData.parse(BufRef(case_serialized))  # noqa: E501
    case_serialized = bytes.fromhex("8001f0970c780300")  # overflow
    assert SmpmDlDeviceEnergy8BGetDataData(year=2000, month=SmpmDlDeviceEnergy8bGetDataDataRequestMonth.DEC, day=31, request_data_pack_ids=(SmpmDlDeviceEnergy8bGetDataDataRequestId.DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_2, SmpmDlDeviceEnergy8bGetDataDataRequestId.NETWORK_PARAMS_PHASE1)) == SmpmDlDeviceEnergy8BGetDataData.parse(BufRef(case_serialized))  # noqa: E501
    case_serialized = bytes.fromhex("80f9f3970c780300")  # overflow
    assert SmpmDlDeviceEnergy8BGetDataData(year=2127, month=SmpmDlDeviceEnergy8bGetDataDataRequestMonth.DEC, day=31, request_data_pack_ids=(SmpmDlDeviceEnergy8bGetDataDataRequestId.DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_2, SmpmDlDeviceEnergy8bGetDataDataRequestId.NETWORK_PARAMS_PHASE1)) == SmpmDlDeviceEnergy8BGetDataData.parse(BufRef(case_serialized))  # noqa: E501



from datetime import timedelta, time, datetime, tzinfo
from enum import IntEnum, unique
from typing import Dict, Tuple, List, Any, Optional

from data_aggregator_sdk.integration_message import ProfileKind, IntegrationV0MessageData, GRANULATION_TO_END_OF_DATETIME_MAP, ProfileGranulation, IntegrationV0MessageProfile
from data_aggregator_sdk.utils.round_dt import round_dt

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.days_ago_calculation import days_ago_calculation
from data_gateway_sdk.utils.packet import Packet


# PACKET (128 bits)   smpm_ul_device_energy_16b_profile_8h_energy
#
# RESULT int:        290799050787508182044066107677279152
# RESULT bin:  MSB   00000000 00111000 00000001 10000000 00001010 00000000 01000000 00000001 10000000 00001000 00000000 00100000 00000000 01010101 00000011 10110000   LSB
# RESULT hex:  LE    b0 03 55 00 20 00 08 80 01 40 00 0a 80 01 38 00
#
# name                       type       size  value(int)                                                                                                                        data(bits)  # noqa: E501
# -------------------------  ---------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------  # noqa: E501
# packet_type_id_enum.0.VAL  u7            7          48                                                                                                                           0110000  # noqa: E501
# packet_type_id_enum.0.DFF  bool          1           1                                                                                                                          1
# packet_type_id_enum.1.VAL  u2            2           3                                                                                                                        11
# packet_type_id_enum.1.DFF  bool          1           0                                                                                                                       0
# days_ago                   timedelta     5           0                                                                                                                  00000
# valid.0.value_valid        bool          1           1                                                                                                                 1
# valid.1.value_valid        bool          1           0                                                                                                                0
# valid.2.value_valid        bool          1           1                                                                                                               1
# valid.3.value_valid        bool          1           0                                                                                                              0
# valid.4.value_valid        bool          1           1                                                                                                             1
# valid.5.value_valid        bool          1           0                                                                                                            0
# valid.6.value_valid        bool          1           1                                                                                                           1
# valid.7.value_valid        bool          1           0                                                                                                          0
# profile.0.point            u13          13           0                                                                                             0000000000000
# profile.1.point            u13          13           1                                                                                0000000000001
# profile.2.point            u13          13           2                                                                   0000000000010
# profile.3.point            u13          13           3                                                      0000000000011
# profile.4.point            u13          13           4                                         0000000000100
# profile.5.point            u13          13           5                            0000000000101
# profile.6.point            u13          13           6               0000000000110
# profile.7.point            u13          13           7  0000000000111

@unique
class SmpmUlDeviceEnergy16bProfile8hEnergyIds(IntEnum):
    UL_DATA_16B__PROFILE_ENERGY_ACTIVE_CONSUMED_H8_1 = 432
    UL_DATA_16B__PROFILE_ENERGY_ACTIVE_CONSUMED_H8_2 = 433
    UL_DATA_16B__PROFILE_ENERGY_ACTIVE_CONSUMED_H8_3 = 434
    UL_DATA_16B__PROFILE_ENERGY_ACTIVE_GENERATED_H8_1 = 435
    UL_DATA_16B__PROFILE_ENERGY_ACTIVE_GENERATED_H8_2 = 436
    UL_DATA_16B__PROFILE_ENERGY_ACTIVE_GENERATED_H8_3 = 437
    UL_DATA_16B__PROFILE_ENERGY_REACTIVE_CONSUMED_H8_1 = 438
    UL_DATA_16B__PROFILE_ENERGY_REACTIVE_CONSUMED_H8_2 = 439
    UL_DATA_16B__PROFILE_ENERGY_REACTIVE_CONSUMED_H8_3 = 440
    UL_DATA_16B__PROFILE_ENERGY_REACTIVE_GENERATED_H8_1 = 441
    UL_DATA_16B__PROFILE_ENERGY_REACTIVE_GENERATED_H8_2 = 442
    UL_DATA_16B__PROFILE_ENERGY_REACTIVE_GENERATED_H8_3 = 443


PROFILE_H8_ENERGY_MAP = {
    SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_ACTIVE_CONSUMED_H8_1: ProfileKind.ENERGY_A_P_DELTA,
    SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_ACTIVE_CONSUMED_H8_2: ProfileKind.ENERGY_A_P_DELTA,
    SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_ACTIVE_CONSUMED_H8_3: ProfileKind.ENERGY_A_P_DELTA,
    SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_ACTIVE_GENERATED_H8_1: ProfileKind.ENERGY_A_N_DELTA,
    SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_ACTIVE_GENERATED_H8_2: ProfileKind.ENERGY_A_N_DELTA,
    SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_ACTIVE_GENERATED_H8_3: ProfileKind.ENERGY_A_N_DELTA,
    SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_REACTIVE_CONSUMED_H8_1: ProfileKind.ENERGY_R_P_DELTA,
    SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_REACTIVE_CONSUMED_H8_2: ProfileKind.ENERGY_R_P_DELTA,
    SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_REACTIVE_CONSUMED_H8_3: ProfileKind.ENERGY_R_P_DELTA,
    SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_REACTIVE_GENERATED_H8_1: ProfileKind.ENERGY_R_N_DELTA,
    SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_REACTIVE_GENERATED_H8_2: ProfileKind.ENERGY_R_N_DELTA,
    SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_REACTIVE_GENERATED_H8_3: ProfileKind.ENERGY_R_N_DELTA,
}


class SmpmUlDeviceEnergy16BProfile8HEnergyData(Packet):
    packet_type_id_enum: SmpmUlDeviceEnergy16bProfile8hEnergyIds
    days_ago: timedelta
    valid: Tuple[bool, bool, bool, bool, bool, bool, bool, bool]
    profile: Tuple[int, int, int, int, int, int, int, int]

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        packet_type_id_enum__value_int_tmp1 = 0
        packet_type_id_enum__value_size_tmp2 = 0
        assert isinstance(data.packet_type_id_enum, SmpmUlDeviceEnergy16bProfile8hEnergyIds)
        packet_type_id_enum__value_int_tmp1 |= ((data.packet_type_id_enum.value) & (2 ** (11) - 1)) << packet_type_id_enum__value_size_tmp2
        packet_type_id_enum__value_size_tmp2 += 11
        packet_type_id_enum__max_size_tmp3 = 0
        packet_type_id_enum__steps_tmp4 = [8, 3, 3, 3]
        for packet_type_id_enum__j_tmp5 in range(32):
            packet_type_id_enum__step_tmp6 = packet_type_id_enum__steps_tmp4[packet_type_id_enum__j_tmp5] if packet_type_id_enum__j_tmp5 < len(packet_type_id_enum__steps_tmp4) else packet_type_id_enum__steps_tmp4[-1]
            packet_type_id_enum__max_size_tmp3 += packet_type_id_enum__step_tmp6
            packet_type_id_enum__current_part_value_tmp7 = packet_type_id_enum__value_int_tmp1 & (2 ** packet_type_id_enum__step_tmp6 - 1)
            packet_type_id_enum__value_int_tmp1 = packet_type_id_enum__value_int_tmp1 >> (packet_type_id_enum__step_tmp6 - 1)
            result |= ((packet_type_id_enum__current_part_value_tmp7) & (2 ** ((packet_type_id_enum__step_tmp6 - 1)) - 1)) << size
            size += (packet_type_id_enum__step_tmp6 - 1)
            assert isinstance((packet_type_id_enum__value_int_tmp1 != 0), bool)
            result |= ((int((packet_type_id_enum__value_int_tmp1 != 0))) & (2 ** (1) - 1)) << size
            size += 1
            if packet_type_id_enum__value_int_tmp1 == 0:
                break
        isinstance(data.days_ago, (int, timedelta))
        days_ago_tmp8 = int(data.days_ago.total_seconds() // 86400 if isinstance(data.days_ago, timedelta) else data.days_ago // 86400)
        assert 0 <= days_ago_tmp8 <= 31
        result |= ((days_ago_tmp8) & (2 ** (5) - 1)) << size
        size += 5
        assert isinstance(data.valid, tuple) and len(data.valid) == 8
        assert isinstance(data.valid[0], bool)
        result |= ((int(data.valid[0])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.valid[1], bool)
        result |= ((int(data.valid[1])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.valid[2], bool)
        result |= ((int(data.valid[2])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.valid[3], bool)
        result |= ((int(data.valid[3])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.valid[4], bool)
        result |= ((int(data.valid[4])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.valid[5], bool)
        result |= ((int(data.valid[5])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.valid[6], bool)
        result |= ((int(data.valid[6])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.valid[7], bool)
        result |= ((int(data.valid[7])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.profile, tuple) and len(data.profile) == 8
        assert isinstance(data.profile[0], int)
        assert 0 <= data.profile[0] <= 8191
        result |= ((data.profile[0]) & (2 ** (13) - 1)) << size
        size += 13
        assert isinstance(data.profile[1], int)
        assert 0 <= data.profile[1] <= 8191
        result |= ((data.profile[1]) & (2 ** (13) - 1)) << size
        size += 13
        assert isinstance(data.profile[2], int)
        assert 0 <= data.profile[2] <= 8191
        result |= ((data.profile[2]) & (2 ** (13) - 1)) << size
        size += 13
        assert isinstance(data.profile[3], int)
        assert 0 <= data.profile[3] <= 8191
        result |= ((data.profile[3]) & (2 ** (13) - 1)) << size
        size += 13
        assert isinstance(data.profile[4], int)
        assert 0 <= data.profile[4] <= 8191
        result |= ((data.profile[4]) & (2 ** (13) - 1)) << size
        size += 13
        assert isinstance(data.profile[5], int)
        assert 0 <= data.profile[5] <= 8191
        result |= ((data.profile[5]) & (2 ** (13) - 1)) << size
        size += 13
        assert isinstance(data.profile[6], int)
        assert 0 <= data.profile[6] <= 8191
        result |= ((data.profile[6]) & (2 ** (13) - 1)) << size
        size += 13
        assert isinstance(data.profile[7], int)
        assert 0 <= data.profile[7] <= 8191
        result |= ((data.profile[7]) & (2 ** (13) - 1)) << size
        size += 13
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceEnergy16BProfile8HEnergyData':
        result__el_tmp9: Dict[str, Any] = dict()
        packet_type_id_enum__res_tmp10 = 0
        packet_type_id_enum__steps_tmp11 = (8, 3, 3, 3)
        packet_type_id_enum__res_size_tmp12 = 0
        packet_type_id_enum__step_tmp13 = 0
        for packet_type_id_enum__i_tmp15 in range(32):
            packet_type_id_enum__step_tmp13 = (packet_type_id_enum__steps_tmp11[packet_type_id_enum__i_tmp15] if packet_type_id_enum__i_tmp15 < len(
                packet_type_id_enum__steps_tmp11) else packet_type_id_enum__steps_tmp11[-1]) - 1
            packet_type_id_enum__res_tmp10 |= buf.shift(packet_type_id_enum__step_tmp13) << packet_type_id_enum__res_size_tmp12
            packet_type_id_enum__res_size_tmp12 += packet_type_id_enum__step_tmp13
            packet_type_id_enum__dff_tmp14 = bool(buf.shift(1))
            if not packet_type_id_enum__dff_tmp14:
                break
        packet_type_id_enum__buf_tmp16 = buf
        buf = BufRef(packet_type_id_enum__res_tmp10, stop_on_buffer_end=True)
        result__el_tmp9["packet_type_id_enum"] = SmpmUlDeviceEnergy16bProfile8hEnergyIds(buf.shift(11))
        buf = packet_type_id_enum__buf_tmp16
        result__el_tmp9["days_ago"] = timedelta(seconds=buf.shift(5) * 86400)
        valid_tmp17: List[bool] = []
        valid__item_tmp18 = bool(buf.shift(1))
        valid_tmp17.append(valid__item_tmp18)
        valid__item_tmp18 = bool(buf.shift(1))
        valid_tmp17.append(valid__item_tmp18)
        valid__item_tmp18 = bool(buf.shift(1))
        valid_tmp17.append(valid__item_tmp18)
        valid__item_tmp18 = bool(buf.shift(1))
        valid_tmp17.append(valid__item_tmp18)
        valid__item_tmp18 = bool(buf.shift(1))
        valid_tmp17.append(valid__item_tmp18)
        valid__item_tmp18 = bool(buf.shift(1))
        valid_tmp17.append(valid__item_tmp18)
        valid__item_tmp18 = bool(buf.shift(1))
        valid_tmp17.append(valid__item_tmp18)
        valid__item_tmp18 = bool(buf.shift(1))
        valid_tmp17.append(valid__item_tmp18)
        result__el_tmp9["valid"] = tuple(valid_tmp17)
        profile_tmp19: List[int] = []
        profile__item_tmp20 = buf.shift(13) + 0
        profile_tmp19.append(profile__item_tmp20)
        profile__item_tmp20 = buf.shift(13) + 0
        profile_tmp19.append(profile__item_tmp20)
        profile__item_tmp20 = buf.shift(13) + 0
        profile_tmp19.append(profile__item_tmp20)
        profile__item_tmp20 = buf.shift(13) + 0
        profile_tmp19.append(profile__item_tmp20)
        profile__item_tmp20 = buf.shift(13) + 0
        profile_tmp19.append(profile__item_tmp20)
        profile__item_tmp20 = buf.shift(13) + 0
        profile_tmp19.append(profile__item_tmp20)
        profile__item_tmp20 = buf.shift(13) + 0
        profile_tmp19.append(profile__item_tmp20)
        profile__item_tmp20 = buf.shift(13) + 0
        profile_tmp19.append(profile__item_tmp20)
        result__el_tmp9["profile"] = tuple(profile_tmp19)
        result = SmpmUlDeviceEnergy16BProfile8HEnergyData(**result__el_tmp9)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        values: Tuple[Optional[float], ...]
        profiles = (
            self.profile[0] if self.valid[0] else None,
            self.profile[1] if self.valid[1] else None,
            self.profile[2] if self.valid[2] else None,
            self.profile[3] if self.valid[3] else None,
            self.profile[4] if self.valid[4] else None,
            self.profile[5] if self.valid[5] else None,
            self.profile[6] if self.valid[6] else None,
            self.profile[7] if self.valid[7] else None,
        )
        if self.packet_type_id_enum in {432, 435, 438, 441}:
            values = (
                *profiles,
                None, None, None, None, None, None, None, None,
                None, None, None, None, None, None, None, None,
            )
        elif self.packet_type_id_enum in {433, 436, 439, 442}:
            values = (
                None, None, None, None, None, None, None, None,
                *profiles,
                None, None, None, None, None, None, None, None,
            )
        else:
            values = (
                None, None, None, None, None, None, None, None,
                None, None, None, None, None, None, None, None,
                *profiles,
            )
        return [
            IntegrationV0MessageData(
                dt=round_dt(
                    days_ago_calculation(
                        received_at,
                        device_tz,
                        time(0),
                        self.days_ago,
                    ),
                    GRANULATION_TO_END_OF_DATETIME_MAP[ProfileGranulation.MINUTE_60],
                ),
                profiles=[
                    IntegrationV0MessageProfile(
                        type=PROFILE_H8_ENERGY_MAP[self.packet_type_id_enum],
                        granulation=ProfileGranulation.MINUTE_60,
                        values=values,
                    ),
                ],
            ),
        ]



from datetime import timedelta

from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_journal import SmpmUlDeviceEnergy16BJournalData, \
    EventData, SmpmUlDeviceEnergy16bJournalType
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_energy_16b_journal() -> None:
    case_serialized = bytes.fromhex("7300cc38000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BJournalData(days_ago=timedelta(seconds=0.0), valid=False, time_offset=timedelta(seconds=14400.0), journal=(EventData(offset=timedelta(seconds=180.0), code=SmpmUlDeviceEnergy16bJournalType.CHANGE_OFFSET_DAILY_CLOCK), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE))) == SmpmUlDeviceEnergy16BJournalData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BJournalData.serialize(SmpmUlDeviceEnergy16BJournalData(days_ago=timedelta(seconds=0.0), valid=False, time_offset=timedelta(seconds=14400.0), journal=(EventData(offset=timedelta(seconds=180.0), code=SmpmUlDeviceEnergy16bJournalType.CHANGE_OFFSET_DAILY_CLOCK), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE))))  # noqa: E501
    case_serialized = bytes.fromhex("7380c038000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BJournalData(days_ago=timedelta(seconds=0.0), valid=True, time_offset=timedelta(seconds=0.0), journal=(EventData(offset=timedelta(seconds=180.0), code=SmpmUlDeviceEnergy16bJournalType.CHANGE_OFFSET_DAILY_CLOCK), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE))) == SmpmUlDeviceEnergy16BJournalData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BJournalData.serialize(SmpmUlDeviceEnergy16BJournalData(days_ago=timedelta(seconds=0.0), journal=(EventData(offset=timedelta(seconds=180.0), code=SmpmUlDeviceEnergy16bJournalType.CHANGE_OFFSET_DAILY_CLOCK), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE)), time_offset=timedelta(seconds=0.0), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("7380ff38000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BJournalData(days_ago=timedelta(seconds=0.0), valid=True, time_offset=timedelta(seconds=75600.0), journal=(EventData(offset=timedelta(seconds=180.0), code=SmpmUlDeviceEnergy16bJournalType.CHANGE_OFFSET_DAILY_CLOCK), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE))) == SmpmUlDeviceEnergy16BJournalData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BJournalData.serialize(SmpmUlDeviceEnergy16BJournalData(days_ago=timedelta(seconds=0.0), journal=(EventData(offset=timedelta(seconds=180.0), code=SmpmUlDeviceEnergy16bJournalType.CHANGE_OFFSET_DAILY_CLOCK), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE)), time_offset=timedelta(seconds=75600.0), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("73ffc038000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BJournalData(days_ago=timedelta(seconds=10972800.0), valid=True, time_offset=timedelta(seconds=0.0), journal=(EventData(offset=timedelta(seconds=180.0), code=SmpmUlDeviceEnergy16bJournalType.CHANGE_OFFSET_DAILY_CLOCK), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE))) == SmpmUlDeviceEnergy16BJournalData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BJournalData.serialize(SmpmUlDeviceEnergy16BJournalData(days_ago=timedelta(seconds=10972800.0), journal=(EventData(offset=timedelta(seconds=180.0), code=SmpmUlDeviceEnergy16bJournalType.CHANGE_OFFSET_DAILY_CLOCK), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE)), time_offset=timedelta(seconds=0.0), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("73ffff38000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BJournalData(days_ago=timedelta(seconds=10972800.0), valid=True, time_offset=timedelta(seconds=75600.0), journal=(EventData(offset=timedelta(seconds=180.0), code=SmpmUlDeviceEnergy16bJournalType.CHANGE_OFFSET_DAILY_CLOCK), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE))) == SmpmUlDeviceEnergy16BJournalData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BJournalData.serialize(SmpmUlDeviceEnergy16BJournalData(days_ago=timedelta(seconds=10972800.0), journal=(EventData(offset=timedelta(seconds=180.0), code=SmpmUlDeviceEnergy16bJournalType.CHANGE_OFFSET_DAILY_CLOCK), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE), EventData(offset=timedelta(seconds=0.0), code=SmpmUlDeviceEnergy16bJournalType.NONE)), time_offset=timedelta(seconds=75600.0), valid=True))  # noqa: E501



from datetime import timedelta

from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_info import SmpmUlDeviceEnergy16BInfoData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_energy_16b_info() -> None:
    case_serialized = bytes.fromhex("bc0aed017c5687000000000000000000")
    assert SmpmUlDeviceEnergy16BInfoData(battery_volts=3.3, temperature=23, date_time=timedelta(seconds=567648000.0), relay_is_active=False) == SmpmUlDeviceEnergy16BInfoData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BInfoData.serialize(SmpmUlDeviceEnergy16BInfoData(battery_volts=3.3, temperature=23, date_time=timedelta(seconds=567648000.0), relay_is_active=False))  # noqa: E501
    case_serialized = bytes.fromhex("bc02fcffffffff030000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BInfoData(battery_volts=0.0, temperature=155, date_time=timedelta(seconds=2147483647.0), relay_is_active=True) == SmpmUlDeviceEnergy16BInfoData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BInfoData.serialize(SmpmUlDeviceEnergy16BInfoData(battery_volts=0.0, date_time=timedelta(seconds=2147483647.0), relay_is_active=True, temperature=155))  # noqa: E501
    case_serialized = bytes.fromhex("bc0200fcffffff030000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BInfoData(battery_volts=0.0, temperature=-100, date_time=timedelta(seconds=2147483647.0), relay_is_active=True) == SmpmUlDeviceEnergy16BInfoData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BInfoData.serialize(SmpmUlDeviceEnergy16BInfoData(battery_volts=0.0, date_time=timedelta(seconds=2147483647.0), relay_is_active=True, temperature=-100))  # noqa: E501
    case_serialized = bytes.fromhex("bc02fc03000000020000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BInfoData(battery_volts=0.0, temperature=155, date_time=timedelta(seconds=0.0), relay_is_active=True) == SmpmUlDeviceEnergy16BInfoData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BInfoData.serialize(SmpmUlDeviceEnergy16BInfoData(battery_volts=0.0, date_time=timedelta(seconds=0.0), relay_is_active=True, temperature=155))  # noqa: E501
    case_serialized = bytes.fromhex("bc020000000000020000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BInfoData(battery_volts=0.0, temperature=-100, date_time=timedelta(seconds=0.0), relay_is_active=True) == SmpmUlDeviceEnergy16BInfoData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BInfoData.serialize(SmpmUlDeviceEnergy16BInfoData(battery_volts=0.0, date_time=timedelta(seconds=0.0), relay_is_active=True, temperature=-100))  # noqa: E501
    case_serialized = bytes.fromhex("bcfaffffffffff030000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BInfoData(battery_volts=12.7, temperature=155, date_time=timedelta(seconds=2147483647.0), relay_is_active=True) == SmpmUlDeviceEnergy16BInfoData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BInfoData.serialize(SmpmUlDeviceEnergy16BInfoData(battery_volts=12.7, date_time=timedelta(seconds=2147483647.0), relay_is_active=True, temperature=155))  # noqa: E501
    case_serialized = bytes.fromhex("bcfa03fcffffff030000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BInfoData(battery_volts=12.7, temperature=-100, date_time=timedelta(seconds=2147483647.0), relay_is_active=True) == SmpmUlDeviceEnergy16BInfoData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BInfoData.serialize(SmpmUlDeviceEnergy16BInfoData(battery_volts=12.7, date_time=timedelta(seconds=2147483647.0), relay_is_active=True, temperature=-100))  # noqa: E501
    case_serialized = bytes.fromhex("bcfaff03000000020000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BInfoData(battery_volts=12.7, temperature=155, date_time=timedelta(seconds=0.0), relay_is_active=True) == SmpmUlDeviceEnergy16BInfoData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BInfoData.serialize(SmpmUlDeviceEnergy16BInfoData(battery_volts=12.7, date_time=timedelta(seconds=0.0), relay_is_active=True, temperature=155))  # noqa: E501
    case_serialized = bytes.fromhex("bcfa0300000000020000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BInfoData(battery_volts=12.7, temperature=-100, date_time=timedelta(seconds=0.0), relay_is_active=True) == SmpmUlDeviceEnergy16BInfoData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BInfoData.serialize(SmpmUlDeviceEnergy16BInfoData(battery_volts=12.7, date_time=timedelta(seconds=0.0), relay_is_active=True, temperature=-100))  # noqa: E501



from datetime import timedelta, datetime, time, tzinfo
from enum import IntEnum, unique
from typing import List, Any, Dict, Tuple

import pytz
from data_aggregator_sdk.constants.enums import DeviceHack
from data_aggregator_sdk.utils.round_dt import round_dt
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, \
    IntegrationV0MessageProfile, ProfileKind, ProfileGranulation, GRANULATION_TO_END_OF_DATETIME_MAP

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.days_ago_calculation import days_ago_calculation
from data_gateway_sdk.utils.packet import Packet
from data_gateway_sdk.utils.true_round import true_round


# PACKET (128 bits)   smpm_ul_device_energy_16b_profile_8h2_energy
#
# RESULT int:        265850142892151729396105850589001287275
# RESULT bin:  MSB   11001000 00000000 11100000 00000110 00000000 00101000 00000001 00000000 00000110 00000000 00100000 00000000 10000000 00000000 00000110 01101011   LSB
# RESULT hex:  LE    6b 06 00 80 00 20 00 06 00 01 28 00 06 e0 00 c8
#
# name                     type                                       size  value(int)                                                                                                                        data(bits)    # noqa: E501
# -----------------------  -----------------------------------------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------    # noqa: E501
# packet_type_id.0.VAL     u7                                            7         107                                                                                                                           1101011    # noqa: E501
# packet_type_id.0.DFF     bool                                          1           0                                                                                                                          0           # noqa: E501
# type                     SmpmUlDeviceEnergy16bProfile8h2EnergyType     2           2                                                                                                                        10            # noqa: E501
# point_factor_multiplier  u2                                            2           1                                                                                                                      01              # noqa: E501
# days_ago                 timedelta                                     6           0                                                                                                                000000                # noqa: E501
# profile.0.point          u13                                          13           0                                                                                                   0000000000000                      # noqa: E501
# profile.1.point          u13                                          13           1                                                                                      0000000000001                                   # noqa: E501
# profile.2.point          u13                                          13           2                                                                         0000000000010
# profile.3.point          u13                                          13           3                                                            0000000000011
# profile.4.point          u13                                          13           4                                               0000000000100
# profile.5.point          u13                                          13           5                                  0000000000101
# profile.6.point          u13                                          13           6                     0000000000110
# profile.7.point          u13                                          13           7        0000000000111
# point_factor             uf6p1                                         6          50  110010

@unique
class SmpmUlDeviceEnergy16bProfile8h2EnergyType(IntEnum):
    ENERGY_GENERATED_ACTIVE = 0  # ENERGY_A_N 0
    ENERGY_GENERATED_REACTIVE = 1  # ENERGY_R_N 1
    ENERGY_CONSUMED_ACTIVE = 2  # ENERGY_A_P 2
    ENERGY_CONSUMED_REACTIVE = 3  # ENERGY_R_P 3

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


PROFILE_MAP = {
    SmpmUlDeviceEnergy16bProfile8h2EnergyType.ENERGY_GENERATED_ACTIVE: ProfileKind.ENERGY_A_N_DELTA,
    SmpmUlDeviceEnergy16bProfile8h2EnergyType.ENERGY_GENERATED_REACTIVE: ProfileKind.ENERGY_R_N_DELTA,
    SmpmUlDeviceEnergy16bProfile8h2EnergyType.ENERGY_CONSUMED_ACTIVE: ProfileKind.ENERGY_A_P_DELTA,
    SmpmUlDeviceEnergy16bProfile8h2EnergyType.ENERGY_CONSUMED_REACTIVE: ProfileKind.ENERGY_R_P_DELTA,
}


class SmpmUlDeviceEnergy16BProfile8H2EnergyData(Packet):
    type: SmpmUlDeviceEnergy16bProfile8h2EnergyType
    point_factor_multiplier: int
    days_ago: timedelta
    profile: Tuple[int, int, int, int, int, int, int, int]
    point_factor: float

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((107) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.type, SmpmUlDeviceEnergy16bProfile8h2EnergyType)
        result |= ((data.type.value) & (2 ** (2) - 1)) << size
        size += 2
        assert isinstance(data.point_factor_multiplier, int)
        assert 1 <= data.point_factor_multiplier <= 4
        result |= (((data.point_factor_multiplier + -1)) & (2 ** (2) - 1)) << size
        size += 2
        isinstance(data.days_ago, (int, timedelta))
        days_ago_tmp1 = int(data.days_ago.total_seconds() // 86400 if isinstance(data.days_ago, timedelta) else data.days_ago // 86400)
        assert 0 <= days_ago_tmp1 <= 63
        result |= ((days_ago_tmp1) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.profile, tuple) and len(data.profile) == 8
        assert isinstance(data.profile[0], int)
        assert 0 <= data.profile[0] <= 8191
        result |= ((data.profile[0]) & (2 ** (13) - 1)) << size
        size += 13
        assert isinstance(data.profile[1], int)
        assert 0 <= data.profile[1] <= 8191
        result |= ((data.profile[1]) & (2 ** (13) - 1)) << size
        size += 13
        assert isinstance(data.profile[2], int)
        assert 0 <= data.profile[2] <= 8191
        result |= ((data.profile[2]) & (2 ** (13) - 1)) << size
        size += 13
        assert isinstance(data.profile[3], int)
        assert 0 <= data.profile[3] <= 8191
        result |= ((data.profile[3]) & (2 ** (13) - 1)) << size
        size += 13
        assert isinstance(data.profile[4], int)
        assert 0 <= data.profile[4] <= 8191
        result |= ((data.profile[4]) & (2 ** (13) - 1)) << size
        size += 13
        assert isinstance(data.profile[5], int)
        assert 0 <= data.profile[5] <= 8191
        result |= ((data.profile[5]) & (2 ** (13) - 1)) << size
        size += 13
        assert isinstance(data.profile[6], int)
        assert 0 <= data.profile[6] <= 8191
        result |= ((data.profile[6]) & (2 ** (13) - 1)) << size
        size += 13
        assert isinstance(data.profile[7], int)
        assert 0 <= data.profile[7] <= 8191
        result |= ((data.profile[7]) & (2 ** (13) - 1)) << size
        size += 13
        assert isinstance(data.point_factor, (int, float))
        assert 0.0 <= data.point_factor <= 6.3
        result |= ((int(round(float(data.point_factor) * 10.0, 0))) & (2 ** (6) - 1)) << size
        size += 6
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceEnergy16BProfile8H2EnergyData':
        result__el_tmp2: Dict[str, Any] = dict()
        if 107 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp2["type"] = SmpmUlDeviceEnergy16bProfile8h2EnergyType(buf.shift(2))
        result__el_tmp2["point_factor_multiplier"] = buf.shift(2) + 1
        result__el_tmp2["days_ago"] = timedelta(seconds=buf.shift(6) * 86400)
        profile_tmp3: List[int] = []
        profile__item_tmp4 = buf.shift(13) + 0
        profile_tmp3.append(profile__item_tmp4)
        profile__item_tmp4 = buf.shift(13) + 0
        profile_tmp3.append(profile__item_tmp4)
        profile__item_tmp4 = buf.shift(13) + 0
        profile_tmp3.append(profile__item_tmp4)
        profile__item_tmp4 = buf.shift(13) + 0
        profile_tmp3.append(profile__item_tmp4)
        profile__item_tmp4 = buf.shift(13) + 0
        profile_tmp3.append(profile__item_tmp4)
        profile__item_tmp4 = buf.shift(13) + 0
        profile_tmp3.append(profile__item_tmp4)
        profile__item_tmp4 = buf.shift(13) + 0
        profile_tmp3.append(profile__item_tmp4)
        profile__item_tmp4 = buf.shift(13) + 0
        profile_tmp3.append(profile__item_tmp4)
        result__el_tmp2["profile"] = tuple(profile_tmp3)
        result__el_tmp2["point_factor"] = round(buf.shift(6) / 10.0, 1)
        result = SmpmUlDeviceEnergy16BProfile8H2EnergyData(**result__el_tmp2)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        point_factor = self.point_factor * self.point_factor_multiplier
        hacks = kwargs.get('hacks', [])
        if DeviceHack.electricity_profile_packets_days_ago_is_zero in hacks:
            return [
                IntegrationV0MessageData(
                    dt=(
                        round_dt(
                            received_at.replace(tzinfo=pytz.UTC).astimezone(device_tz),
                            GRANULATION_TO_END_OF_DATETIME_MAP[ProfileGranulation.MINUTE_60],
                        ) - timedelta(seconds=3600 * 8)).astimezone(pytz.timezone('UTC')),
                    profiles=[
                        IntegrationV0MessageProfile(
                            type=PROFILE_MAP[self.type],
                            granulation=ProfileGranulation.MINUTE_60,
                            values=(
                                true_round(self.profile[0] * point_factor) / 1000,
                                true_round(self.profile[1] * point_factor) / 1000,
                                true_round(self.profile[2] * point_factor) / 1000,
                                true_round(self.profile[3] * point_factor) / 1000,
                                true_round(self.profile[4] * point_factor) / 1000,
                                true_round(self.profile[5] * point_factor) / 1000,
                                true_round(self.profile[6] * point_factor) / 1000,
                                true_round(self.profile[7] * point_factor) / 1000,
                                None, None, None, None, None, None, None, None,
                                None, None, None, None, None, None, None, None,
                            ),
                        ),
                    ],
                ),
            ]
        return [
            IntegrationV0MessageData(
                dt=round_dt(
                    days_ago_calculation(
                        received_at,
                        device_tz,
                        time(0),
                        self.days_ago,
                    ),
                    GRANULATION_TO_END_OF_DATETIME_MAP[ProfileGranulation.MINUTE_60],
                ),
                profiles=[
                    IntegrationV0MessageProfile(
                        type=PROFILE_MAP[self.type],
                        granulation=ProfileGranulation.MINUTE_60,
                        values=(
                            None, None, None, None, None, None, None, None,
                            true_round(self.profile[0] * point_factor, 3) / 1000,
                            true_round(self.profile[1] * point_factor, 3) / 1000,
                            true_round(self.profile[2] * point_factor, 3) / 1000,
                            true_round(self.profile[3] * point_factor, 3) / 1000,
                            true_round(self.profile[4] * point_factor, 3) / 1000,
                            true_round(self.profile[5] * point_factor, 3) / 1000,
                            true_round(self.profile[6] * point_factor, 3) / 1000,
                            true_round(self.profile[7] * point_factor, 3) / 1000,
                            None, None, None, None, None, None, None, None,
                        ),
                    ),
                ],
            ),
        ]



from decimal import Decimal
from datetime import timedelta, datetime, time, tzinfo
from typing import List, Any, Dict

from data_aggregator_sdk.constants.enums import JournalDataType, SensorType
from data_aggregator_sdk.integration_message import IntegrationV0MessageEvent, IntegrationV0MessageData, \
    IntegrationV0MessageConsumption, CounterType, ResourceType, IntegrationV0MessageClock, IntegrationV0MessageSensor

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.days_ago_calculation import days_ago_calculation
from data_gateway_sdk.utils.packet import Packet
from data_gateway_sdk.utils.timestamp_calculation import timestamp_calculation


# PACKET (128 bits)   smpm_ul_device_jupiter_16b_counter
#
# RESULT int:        170141183381241069172511522525825469569
# RESULT bin:  MSB   01111111 11111111 11111111 11111110 11111111 11111111 11111111 11111101 10010000 10111010 01111111 11111111 11111111 11000000 00001100 10000001   LSB
# RESULT hex:  LE    81 0c c0 ff ff 7f ba 90 fd ff ff ff fe ff ff 7f
#
# name                      type       size  value(int)                                                                                                                        data(bits)   # noqa: E501
# ------------------------  ---------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------   # noqa: E501
# packet_type_id.0.VAL      u7            7           1                                                                                                                           0000001   # noqa: E501
# packet_type_id.0.DFF      bool          1           1                                                                                                                          1
# packet_type_id.1.VAL      u2            2           0                                                                                                                        00
# packet_type_id.1.DFF      bool          1           1                                                                                                                       1
# packet_type_id.2.VAL      u2            2           1                                                                                                                     01
# packet_type_id.2.DFF      bool          1           0                                                                                                                    0
# days_ago                  timedelta     5           0                                                                                                               00000
# sync_time_days_ago        timedelta     3           0                                                                                                            000
# timestamp_s               timedelta    26    33554431                                                                                  01111111111111111111111111
# temperature               u7            7          58                                                                           0111010
# battery_volts             uf6p1         6          33                                                                     100001
# event_reset               bool          1           0                                                                    0
# event_low_battery_level   bool          1           0                                                                   0
# event_temperature_limits  bool          1           1                                                                  1
# event_battery_warn        bool          1           1                                                                 1
# event_system_error        bool          1           0                                                                0
# value_channel_1           u31          31  1073741823                                 0111111111111111111111111111111
# value_channel_2           u31          31  1073741823  0111111111111111111111111111111


class SmpmUlDeviceJupiter16BCounterData(Packet):
    days_ago: timedelta
    sync_time_days_ago: timedelta
    timestamp_s: timedelta
    temperature: int
    battery_volts: float
    event_reset: bool
    event_low_battery_level: bool
    event_temperature_limits: bool
    event_battery_warn: bool
    event_system_error: bool
    value_channel_1: int
    value_channel_2: int

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((1) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((0) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((1) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        isinstance(data.days_ago, (int, timedelta))
        days_ago_tmp1 = int(data.days_ago.total_seconds() // 86400 if isinstance(data.days_ago, timedelta) else data.days_ago // 86400)
        assert 0 <= days_ago_tmp1 <= 31
        result |= ((days_ago_tmp1) & (2 ** (5) - 1)) << size
        size += 5
        isinstance(data.sync_time_days_ago, (int, timedelta))
        result |= ((max(min(7, int(data.sync_time_days_ago.total_seconds() // 86400 if isinstance(data.sync_time_days_ago, timedelta) else data.sync_time_days_ago // 86400)), 0)) & (2 ** (3) - 1)) << size    # noqa: E501
        size += 3
        isinstance(data.timestamp_s, (int, timedelta))
        value_int_tmp2 = int(data.timestamp_s.total_seconds() // 1 if isinstance(data.timestamp_s, timedelta) else data.timestamp_s // 1) & 67108863
        result |= ((value_int_tmp2) & (2 ** (26) - 1)) << size
        size += 26
        assert isinstance(data.temperature, int)
        assert -35 <= data.temperature <= 92
        result |= (((data.temperature + 35)) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.battery_volts, (int, float))
        assert 0.0 <= data.battery_volts <= 6.3
        result |= ((int(round(float(data.battery_volts) * 10.0, 0))) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.event_reset, bool)
        result |= ((int(data.event_reset)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_low_battery_level, bool)
        result |= ((int(data.event_low_battery_level)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_temperature_limits, bool)
        result |= ((int(data.event_temperature_limits)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_battery_warn, bool)
        result |= ((int(data.event_battery_warn)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_system_error, bool)
        result |= ((int(data.event_system_error)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.value_channel_1, int)
        result |= (((data.value_channel_1) & 2147483647) & (2 ** (31) - 1)) << size
        size += 31
        assert isinstance(data.value_channel_2, int)
        result |= (((data.value_channel_2) & 2147483647) & (2 ** (31) - 1)) << size
        size += 31
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceJupiter16BCounterData':
        result__el_tmp3: Dict[str, Any] = dict()
        if 1 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp3["days_ago"] = timedelta(seconds=buf.shift(5) * 86400)
        result__el_tmp3["sync_time_days_ago"] = timedelta(seconds=buf.shift(3) * 86400)
        result__el_tmp3["timestamp_s"] = timedelta(seconds=buf.shift(26) * 1)
        result__el_tmp3["temperature"] = buf.shift(7) + -35
        result__el_tmp3["battery_volts"] = round(buf.shift(6) / 10.0, 1)
        result__el_tmp3["event_reset"] = bool(buf.shift(1))
        result__el_tmp3["event_low_battery_level"] = bool(buf.shift(1))
        result__el_tmp3["event_temperature_limits"] = bool(buf.shift(1))
        result__el_tmp3["event_battery_warn"] = bool(buf.shift(1))
        result__el_tmp3["event_system_error"] = bool(buf.shift(1))
        result__el_tmp3["value_channel_1"] = buf.shift(31) + 0
        result__el_tmp3["value_channel_2"] = buf.shift(31) + 0
        result = SmpmUlDeviceJupiter16BCounterData(**result__el_tmp3)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=days_ago_calculation(received_at, device_tz, time(0, 0, 0), self.days_ago),
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.battery_volts)),
                        sensor_type=SensorType.BATTERY,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.temperature)),
                        sensor_type=SensorType.TEMPERATURE,
                    ),
                ],
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=1,
                        value=Decimal(str(self.value_channel_1)),
                        overloading_value=Decimal(str(2147483647.0)),
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=2,
                        value=Decimal(str(self.value_channel_2)),
                        overloading_value=Decimal(str(2147483647.0)),
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                ],
                events=[
                    *([IntegrationV0MessageEvent.RESET] if self.event_reset else []),
                    *([IntegrationV0MessageEvent.BATTERY_IS_LOW] if self.event_low_battery_level else []),
                    *([IntegrationV0MessageEvent.BATTERY_WARNING] if self.event_battery_warn else []),
                    *([IntegrationV0MessageEvent.TEMPERATURE_LIMIT] if self.event_temperature_limits else []),
                    *([IntegrationV0MessageEvent.ERROR_SYSTEM] if self.event_system_error else []),
                ],
                clock=[IntegrationV0MessageClock(value=timestamp_calculation(received_at, self.timestamp_s, timedelta(seconds=67108863)))] if self.timestamp_s else [],
            ),
        ]



import random
import typing
from enum import IntEnum, unique

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (128 bits)   smpm_dl_device_energy_16b_set_regular_data_sending
#
# RESULT int:        4161395094
# RESULT bin:  MSB   00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 11111000 00001001 11011001 10010110   LSB
# RESULT hex:  LE    96 d9 09 f8 00 00 00 00 00 00 00 00 00 00 00 00
#
# name                  type                                                     size  value(int)                                                                                                                        data(bits)
# --------------------  -------------------------------------------------------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------
# packet_type_id.0.VAL  u7                                                          7          22                                                                                                                           0010110
# packet_type_id.0.DFF  bool                                                        1           1                                                                                                                          1
# packet_type_id.1.VAL  u2                                                          2           1                                                                                                                        01
# packet_type_id.1.DFF  bool                                                        1           0                                                                                                                       0
# pack_id               SmpmDlDeviceEnergy16bSetRegularDataSendingDataRequestId    14         315                                                                                                         00000100111011
# action                SpecUpLinkModelingPacketIdU8Action                          2           0                                                                                                       00
# period_ago            u6                                                          6          31                                                                                                 011111
# week_mask.0.day       bool                                                        1           0                                                                                                0
# week_mask.1.day       bool                                                        1           0                                                                                               0
# week_mask.2.day       bool                                                        1           0                                                                                              0
# week_mask.3.day       bool                                                        1           0                                                                                             0
# week_mask.4.day       bool                                                        1           0                                                                                            0
# week_mask.5.day       bool                                                        1           0                                                                                           0
# week_mask.6.day       bool                                                        1           0                                                                                          0
# days_mask.0.day       bool                                                        1           0                                                                                         0
# days_mask.1.day       bool                                                        1           0                                                                                        0
# days_mask.2.day       bool                                                        1           0                                                                                       0
# days_mask.3.day       bool                                                        1           0                                                                                      0
# days_mask.4.day       bool                                                        1           0                                                                                     0
# days_mask.5.day       bool                                                        1           0                                                                                    0
# days_mask.6.day       bool                                                        1           0                                                                                   0
# days_mask.7.day       bool                                                        1           0                                                                                  0
# days_mask.8.day       bool                                                        1           0                                                                                 0
# days_mask.9.day       bool                                                        1           0                                                                                0
# days_mask.10.day      bool                                                        1           0                                                                               0
# days_mask.11.day      bool                                                        1           0                                                                              0
# days_mask.12.day      bool                                                        1           0                                                                             0
# days_mask.13.day      bool                                                        1           0                                                                            0
# days_mask.14.day      bool                                                        1           0                                                                           0
# days_mask.15.day      bool                                                        1           0                                                                          0
# days_mask.16.day      bool                                                        1           0                                                                         0
# days_mask.17.day      bool                                                        1           0                                                                        0
# days_mask.18.day      bool                                                        1           0                                                                       0
# days_mask.19.day      bool                                                        1           0                                                                      0
# days_mask.20.day      bool                                                        1           0                                                                     0
# days_mask.21.day      bool                                                        1           0                                                                    0
# days_mask.22.day      bool                                                        1           0                                                                   0
# days_mask.23.day      bool                                                        1           0                                                                  0
# days_mask.24.day      bool                                                        1           0                                                                 0
# days_mask.25.day      bool                                                        1           0                                                                0
# days_mask.26.day      bool                                                        1           0                                                               0
# days_mask.27.day      bool                                                        1           0                                                              0
# days_mask.28.day      bool                                                        1           0                                                             0
# days_mask.29.day      bool                                                        1           0                                                            0
# days_mask.30.day      bool                                                        1           0                                                           0
# hour_mask.0.hour      bool                                                        1           0                                                          0
# hour_mask.1.hour      bool                                                        1           0                                                         0
# hour_mask.2.hour      bool                                                        1           0                                                        0
# hour_mask.3.hour      bool                                                        1           0                                                       0
# hour_mask.4.hour      bool                                                        1           0                                                      0
# hour_mask.5.hour      bool                                                        1           0                                                     0
# hour_mask.6.hour      bool                                                        1           0                                                    0
# hour_mask.7.hour      bool                                                        1           0                                                   0
# hour_mask.8.hour      bool                                                        1           0                                                  0
# hour_mask.9.hour      bool                                                        1           0                                                 0
# hour_mask.10.hour     bool                                                        1           0                                                0
# hour_mask.11.hour     bool                                                        1           0                                               0
# hour_mask.12.hour     bool                                                        1           0                                              0
# hour_mask.13.hour     bool                                                        1           0                                             0
# hour_mask.14.hour     bool                                                        1           0                                            0
# hour_mask.15.hour     bool                                                        1           0                                           0
# hour_mask.16.hour     bool                                                        1           0                                          0
# hour_mask.17.hour     bool                                                        1           0                                         0
# hour_mask.18.hour     bool                                                        1           0                                        0
# hour_mask.19.hour     bool                                                        1           0                                       0
# hour_mask.20.hour     bool                                                        1           0                                      0
# hour_mask.21.hour     bool                                                        1           0                                     0
# hour_mask.22.hour     bool                                                        1           0                                    0
# hour_mask.23.hour     bool                                                        1           0                                   0
# RESERVED              u33                                                        33           0  000000000000000000000000000000000
# Reserved space is filled with random bits in DOWNLINK packets !


@unique
class SmpmDlDeviceEnergy16bSetRegularDataSendingDataRequestId(IntEnum):
    UL_DATA_16B__ENERGY = 315
    DAILY_ENERGY_ACTIVE_CONSUMED = 400
    DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_1 = 401
    DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_2 = 402
    DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_3 = 403
    DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_4 = 404
    DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_SUM = 405
    DAILY_ENERGY_REACTIVE_CONSUMED = 406
    DAILY_ENERGY_ACTIVE_GENERATED = 407
    DAILY_ENERGY_REACTIVE_GENERATED = 408
    MONTHLY_ENERGY_ACTIVE_CONSUMED = 409
    MONTHLY_ENERGY_ACTIVE_CONSUMED_TARIFF_1 = 410
    MONTHLY_ENERGY_ACTIVE_CONSUMED_TARIFF_2 = 411
    MONTHLY_ENERGY_ACTIVE_CONSUMED_TARIFF_3 = 412
    MONTHLY_ENERGY_ACTIVE_CONSUMED_TARIFF_4 = 413
    MONTHLY_ENERGY_ACTIVE_CONSUMED_TARIFF_SUM = 414
    MONTHLY_ENERGY_REACTIVE_CONSUMED = 415
    MONTHLY_ENERGY_ACTIVE_GENERATED = 416
    MONTHLY_ENERGY_REACTIVE_GENERATED = 417


@unique
class SpecUpLinkModelingPacketIdU8Action(IntEnum):
    REPLACE = 0
    ADD = 1
    REPLACE_HOUR = 2
    REPLACE_DAY = 3


class SmpmDlDeviceEnergy16BSetRegularDataSendingData(Packet):
    pack_id: SmpmDlDeviceEnergy16bSetRegularDataSendingDataRequestId
    action: SpecUpLinkModelingPacketIdU8Action
    period_ago: int
    week_mask: typing.Tuple[bool, bool, bool, bool, bool, bool, bool]
    days_mask: typing.Tuple[bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool]
    hour_mask: typing.Tuple[bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool, bool]

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((22) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((1) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.pack_id, SmpmDlDeviceEnergy16bSetRegularDataSendingDataRequestId)
        result |= ((data.pack_id.value) & (2 ** (14) - 1)) << size
        size += 14
        assert isinstance(data.action, SpecUpLinkModelingPacketIdU8Action)
        result |= ((data.action.value) & (2 ** (2) - 1)) << size
        size += 2
        assert isinstance(data.period_ago, int)
        assert 0 <= data.period_ago <= 63
        result |= ((data.period_ago) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.week_mask, tuple) and len(data.week_mask) == 7
        assert isinstance(data.week_mask[0], bool)
        result |= ((int(data.week_mask[0])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.week_mask[1], bool)
        result |= ((int(data.week_mask[1])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.week_mask[2], bool)
        result |= ((int(data.week_mask[2])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.week_mask[3], bool)
        result |= ((int(data.week_mask[3])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.week_mask[4], bool)
        result |= ((int(data.week_mask[4])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.week_mask[5], bool)
        result |= ((int(data.week_mask[5])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.week_mask[6], bool)
        result |= ((int(data.week_mask[6])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask, tuple) and len(data.days_mask) == 31
        assert isinstance(data.days_mask[0], bool)
        result |= ((int(data.days_mask[0])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[1], bool)
        result |= ((int(data.days_mask[1])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[2], bool)
        result |= ((int(data.days_mask[2])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[3], bool)
        result |= ((int(data.days_mask[3])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[4], bool)
        result |= ((int(data.days_mask[4])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[5], bool)
        result |= ((int(data.days_mask[5])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[6], bool)
        result |= ((int(data.days_mask[6])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[7], bool)
        result |= ((int(data.days_mask[7])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[8], bool)
        result |= ((int(data.days_mask[8])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[9], bool)
        result |= ((int(data.days_mask[9])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[10], bool)
        result |= ((int(data.days_mask[10])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[11], bool)
        result |= ((int(data.days_mask[11])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[12], bool)
        result |= ((int(data.days_mask[12])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[13], bool)
        result |= ((int(data.days_mask[13])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[14], bool)
        result |= ((int(data.days_mask[14])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[15], bool)
        result |= ((int(data.days_mask[15])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[16], bool)
        result |= ((int(data.days_mask[16])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[17], bool)
        result |= ((int(data.days_mask[17])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[18], bool)
        result |= ((int(data.days_mask[18])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[19], bool)
        result |= ((int(data.days_mask[19])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[20], bool)
        result |= ((int(data.days_mask[20])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[21], bool)
        result |= ((int(data.days_mask[21])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[22], bool)
        result |= ((int(data.days_mask[22])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[23], bool)
        result |= ((int(data.days_mask[23])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[24], bool)
        result |= ((int(data.days_mask[24])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[25], bool)
        result |= ((int(data.days_mask[25])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[26], bool)
        result |= ((int(data.days_mask[26])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[27], bool)
        result |= ((int(data.days_mask[27])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[28], bool)
        result |= ((int(data.days_mask[28])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[29], bool)
        result |= ((int(data.days_mask[29])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.days_mask[30], bool)
        result |= ((int(data.days_mask[30])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask, tuple) and len(data.hour_mask) == 24
        assert isinstance(data.hour_mask[0], bool)
        result |= ((int(data.hour_mask[0])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[1], bool)
        result |= ((int(data.hour_mask[1])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[2], bool)
        result |= ((int(data.hour_mask[2])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[3], bool)
        result |= ((int(data.hour_mask[3])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[4], bool)
        result |= ((int(data.hour_mask[4])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[5], bool)
        result |= ((int(data.hour_mask[5])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[6], bool)
        result |= ((int(data.hour_mask[6])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[7], bool)
        result |= ((int(data.hour_mask[7])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[8], bool)
        result |= ((int(data.hour_mask[8])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[9], bool)
        result |= ((int(data.hour_mask[9])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[10], bool)
        result |= ((int(data.hour_mask[10])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[11], bool)
        result |= ((int(data.hour_mask[11])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[12], bool)
        result |= ((int(data.hour_mask[12])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[13], bool)
        result |= ((int(data.hour_mask[13])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[14], bool)
        result |= ((int(data.hour_mask[14])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[15], bool)
        result |= ((int(data.hour_mask[15])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[16], bool)
        result |= ((int(data.hour_mask[16])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[17], bool)
        result |= ((int(data.hour_mask[17])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[18], bool)
        result |= ((int(data.hour_mask[18])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[19], bool)
        result |= ((int(data.hour_mask[19])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[20], bool)
        result |= ((int(data.hour_mask[20])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[21], bool)
        result |= ((int(data.hour_mask[21])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[22], bool)
        result |= ((int(data.hour_mask[22])) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.hour_mask[23], bool)
        result |= ((int(data.hour_mask[23])) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((random.getrandbits(33)) & (2 ** 33 - 1)) << size
        size += 33
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmDlDeviceEnergy16BSetRegularDataSendingData':
        result__el_tmp1: typing.Dict[str, typing.Any] = dict()
        if 22 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp1["pack_id"] = SmpmDlDeviceEnergy16bSetRegularDataSendingDataRequestId(buf.shift(14))
        result__el_tmp1["action"] = SpecUpLinkModelingPacketIdU8Action(buf.shift(2))
        result__el_tmp1["period_ago"] = buf.shift(6) + 0
        week_mask_tmp2: typing.List[bool] = []
        week_mask__item_tmp3 = bool(buf.shift(1))
        week_mask_tmp2.append(week_mask__item_tmp3)
        week_mask__item_tmp3 = bool(buf.shift(1))
        week_mask_tmp2.append(week_mask__item_tmp3)
        week_mask__item_tmp3 = bool(buf.shift(1))
        week_mask_tmp2.append(week_mask__item_tmp3)
        week_mask__item_tmp3 = bool(buf.shift(1))
        week_mask_tmp2.append(week_mask__item_tmp3)
        week_mask__item_tmp3 = bool(buf.shift(1))
        week_mask_tmp2.append(week_mask__item_tmp3)
        week_mask__item_tmp3 = bool(buf.shift(1))
        week_mask_tmp2.append(week_mask__item_tmp3)
        week_mask__item_tmp3 = bool(buf.shift(1))
        week_mask_tmp2.append(week_mask__item_tmp3)
        result__el_tmp1["week_mask"] = tuple(week_mask_tmp2)
        days_mask_tmp4: typing.List[bool] = []
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        days_mask__item_tmp5 = bool(buf.shift(1))
        days_mask_tmp4.append(days_mask__item_tmp5)
        result__el_tmp1["days_mask"] = tuple(days_mask_tmp4)
        hour_mask_tmp6: typing.List[bool] = []
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        hour_mask__item_tmp7 = bool(buf.shift(1))
        hour_mask_tmp6.append(hour_mask__item_tmp7)
        result__el_tmp1["hour_mask"] = tuple(hour_mask_tmp6)
        result = SmpmDlDeviceEnergy16BSetRegularDataSendingData(**result__el_tmp1)
        buf.shift(33)
        return result



from decimal import Decimal
from datetime import datetime, tzinfo
from typing import Any, Dict, List

from data_aggregator_sdk.constants.enums import IntegrationV0MessageEvent, SensorType
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, IntegrationV0MessageConsumption, CounterType, ResourceType, \
    BatteryId, IntegrationV0MessageSensor

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (128 bits)   smpm_ul_device_heat_proxy_meter_16b_daily
#
# RESULT int:        13453363668248021446286294801998980
# RESULT bin:  MSB   00000000 00000010 10010111 01001101 01010010 10000000 00000000 00000011 00000000 00000000 00001100 11100100 00000000 00000000 01100100 10000100   LSB
# RESULT hex:  LE    84 64 00 00 e4 0c 00 00 03 00 80 52 4d 97 02 00
#
# name                       type    size  value(int)                                                                                                                        data(bits)     # noqa: E501
# -------------------------  ------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------     # noqa: E501
# packet_type_id.0.VAL       u7         7           4                                                                                                                           0000100     # noqa: E501
# packet_type_id.0.DFF       bool       1           1                                                                                                                          1
# packet_type_id.1.VAL       u2         2           0                                                                                                                        00
# packet_type_id.1.DFF       bool       1           1                                                                                                                       1
# packet_type_id.2.VAL       u2         2           0                                                                                                                     00
# packet_type_id.2.DFF       bool       1           1                                                                                                                    1
# packet_type_id.3.VAL       u2         2           1                                                                                                                  01
# packet_type_id.3.DFF       bool       1           0                                                                                                                 0
# UNUSED                     -         15           0                                                                                                  000000000000000
# value                      uf27p3    27        3300                                                                       000000000000000110011100100
# UNUSED                     -          5           0                                                                  00000
# uptime_min                 u22       22           3                                            0000000000000000000011
# meter_battery_volts        uf9p2      9         330                                   101001010
# UNUSED                     -          1           0                                  0
# capacitor_volts            uf9p2      9         333                         101001101
# radio_proxy_battery_volts  uf9p2      9         331                101001011
# error_meter_sync           bool       1           0               0
# error_reset                bool       1           0              0
# RESERVED                   u12       12           0  000000000000


class SmpmUlDeviceHeatProxyMeter16BDailyData(Packet):
    value: float
    uptime_min: int
    meter_battery_volts: float
    capacitor_volts: float
    radio_proxy_battery_volts: float
    error_meter_sync: bool
    error_reset: bool

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((4) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((0) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((0) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((1) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((0) & (2 ** (15) - 1)) << size
        size += 15
        assert isinstance(data.value, (int, float))
        result |= ((int(round(float(data.value) * 1000.0, 0)) & 134217727) & (2 ** (27) - 1)) << size
        size += 27
        result |= ((0) & (2 ** (5) - 1)) << size
        size += 5
        assert isinstance(data.uptime_min, int)
        result |= (((data.uptime_min) & 4194303) & (2 ** (22) - 1)) << size
        size += 22
        assert isinstance(data.meter_battery_volts, (int, float))
        result |= ((int(round(max(min(5.11, float(data.meter_battery_volts)), 0.0) * 100.0, 0))) & (2 ** (9) - 1)) << size
        size += 9
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.capacitor_volts, (int, float))
        result |= ((int(round(max(min(5.11, float(data.capacitor_volts)), 0.0) * 100.0, 0))) & (2 ** (9) - 1)) << size
        size += 9
        assert isinstance(data.radio_proxy_battery_volts, (int, float))
        result |= ((int(round(max(min(5.11, float(data.radio_proxy_battery_volts)), 0.0) * 100.0, 0))) & (2 ** (9) - 1)) << size
        size += 9
        assert isinstance(data.error_meter_sync, bool)
        result |= ((int(data.error_meter_sync)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.error_reset, bool)
        result |= ((int(data.error_reset)) & (2 ** (1) - 1)) << size
        size += 1
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceHeatProxyMeter16BDailyData':
        result__el_tmp1: Dict[str, Any] = dict()
        if 4 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        buf.shift(15)
        result__el_tmp1["value"] = round(buf.shift(27) / 1000.0, 3)
        buf.shift(5)
        result__el_tmp1["uptime_min"] = buf.shift(22) + 0
        result__el_tmp1["meter_battery_volts"] = round(buf.shift(9) / 100.0, 2)
        buf.shift(1)
        result__el_tmp1["capacitor_volts"] = round(buf.shift(9) / 100.0, 2)
        result__el_tmp1["radio_proxy_battery_volts"] = round(buf.shift(9) / 100.0, 2)
        result__el_tmp1["error_meter_sync"] = bool(buf.shift(1))
        result__el_tmp1["error_reset"] = bool(buf.shift(1))
        result = SmpmUlDeviceHeatProxyMeter16BDailyData(**result__el_tmp1)
        buf.shift(12)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        if self.error_meter_sync:
            return [
                IntegrationV0MessageData(
                    dt=received_at,
                    sensors=[
                        IntegrationV0MessageSensor(
                            value=Decimal(str(self.capacitor_volts)),
                            sensor_type=SensorType.BATTERY,
                            sensor_id=BatteryId.CAPACITOR,
                        ),
                        IntegrationV0MessageSensor(
                            value=Decimal(str(self.radio_proxy_battery_volts)),
                            sensor_type=SensorType.BATTERY,
                            sensor_id=BatteryId.RADIO_MODULE,
                        ),
                    ],
                    events=[
                        *([IntegrationV0MessageEvent.ERROR_RESET] if self.error_reset else []),
                        *([IntegrationV0MessageEvent.ERROR_METER_SYNC] if self.error_meter_sync else []),
                    ],
                ),
            ]
        return [
            IntegrationV0MessageData(
                dt=received_at,
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.meter_battery_volts)),
                        sensor_type=SensorType.BATTERY,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.capacitor_volts)),
                        sensor_type=SensorType.BATTERY,
                        sensor_id=BatteryId.CAPACITOR,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.radio_proxy_battery_volts)),
                        sensor_type=SensorType.BATTERY,
                        sensor_id=BatteryId.RADIO_MODULE,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.uptime_min * 60)),
                        sensor_type=SensorType.UPTIME,
                        channel_id=1,
                    ),
                ],
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        value=Decimal(str(self.value)),
                        resource_type=ResourceType.COMMON,
                        channel=1,
                        overloading_value=Decimal(str(134217.727)),
                    ),
                ],
                events=[
                    *([IntegrationV0MessageEvent.ERROR_RESET] if self.error_reset else []),
                    *([IntegrationV0MessageEvent.ERROR_METER_SYNC] if self.error_meter_sync else []),
                ],
            ),
        ]
